import pytest

from paita.rag.models import RAGSource, RAGSources, RAGSourceType
from paita.utils.json_utils import JsonUtils

APP_NAME = "TestApp"
APP_AUTHOR = "TestAuthor"
FILE_NAME = "test_file.json"
RAG_SOURCES = RAGSources(
    rag_sources=[
        RAGSource(
            rag_source_type=RAGSourceType.web_page,
            rag_source="aA",
            rag_source_max_depth=1,
        ),
        RAGSource(
            rag_source_type=RAGSourceType.web_page,
            rag_source="bB",
            rag_source_max_depth=1,
        ),
    ]
)


@pytest.mark.asyncio
async def test_read_not_found():
    instance = JsonUtils(app_name=APP_NAME, app_author=APP_AUTHOR, file_name=FILE_NAME)
    with pytest.raises(FileNotFoundError):
        await instance.read(RAGSources)


@pytest.mark.asyncio
async def test_write_read():
    try:
        instance = JsonUtils(app_name=APP_NAME, app_author=APP_AUTHOR, file_name=FILE_NAME)
        await instance.write(RAG_SOURCES)
        rag_sources = await instance.read(RAGSources)
        assert rag_sources == RAG_SOURCES
    finally:
        await instance.delete()


@pytest.mark.asyncio
async def test_delete_not_found():
    instance = JsonUtils(app_name=APP_NAME, app_author=APP_AUTHOR, file_name=FILE_NAME)
    with pytest.raises(FileNotFoundError):
        await instance.read(RAGSources)


@pytest.mark.asyncio
async def test_delete():
    instance = JsonUtils(app_name=APP_NAME, app_author=APP_AUTHOR, file_name=FILE_NAME)
    await instance.write(RAG_SOURCES)
    await instance.delete()
    with pytest.raises(FileNotFoundError):
        await instance.read(RAGSources)



from __future__ import annotations

from typing import Any, Optional

from pydantic import BaseModel


class SettingsModel(BaseModel):
    version: float = 0.1
    ai_service: Optional[str] = None
    ai_model: Optional[str] = None
    ai_persona: Optional[str] = "You are a helpful assistant. Answer all questions to the best of your ability."
    ai_streaming: Optional[bool] = True
    ai_model_kwargs: Optional[dict[str, Any]] = {}
    ai_n: Optional[int] = 1
    ai_max_tokens: Optional[int] = 2048
    ai_history_depth: Optional[int] = 20

    ai_rag_enabled: Optional[bool] = False
    ai_rag_vector_store_type: Optional[str] = None
    ai_rag_contextualize_prompt: Optional[str] = """Given a chat history and the latest user question \
which might reference context in the chat history, formulate a standalone question \
which can be understood without the chat history. Do NOT answer the question, \
just reformulate it if needed and otherwise return it as is."""
    ai_rag_system_prompt: Optional[str] = """You are an assistant for question-answering tasks. \
Use the following pieces of retrieved context to answer the question. \
If you don't know the answer, just say that you don't know. \
Use two sentences maximum and keep the answer concise.\

{context}"""



import contextlib
from pathlib import Path
from typing import Iterator

from appdirs import user_config_dir
from bs4 import BeautifulSoup as Soup
from langchain.chains import (
    create_history_aware_retriever,
    create_retrieval_chain,
)
from langchain.chains.combine_documents import create_stuff_documents_chain
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders.recursive_url_loader import RecursiveUrlLoader
from langchain_community.vectorstores import Chroma
from langchain_community.vectorstores.utils import filter_complex_metadata
from langchain_core.chat_history import BaseChatMessageHistory
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain_core.vectorstores import VectorStore
from pydantic import BaseModel, ConfigDict

from paita.rag.models import RAGSource, RAGSources, RAGSourceType, RAGVectorStoreType
from paita.utils.json_utils import JsonUtils
from paita.utils.settings_model import SettingsModel


class RAGManagerModel(BaseModel):
    app_name: str
    app_author: str
    embeddings: Embeddings
    vector_store_type: RAGVectorStoreType

    model_config = ConfigDict(arbitrary_types_allowed=True)


class RAGManager:
    def __init__(self, model: RAGManagerModel):
        self.model = model
        config_dir = user_config_dir(appname=self.model.app_name, appauthor=self.model.app_author)
        self.file_path = Path(config_dir) / "rag"
        self.vectorstore = self.create_vectorstore(
            vector_store_type=self.model.vector_store_type,
            persist_directory=str(self.file_path),
            embeddings=self.model.embeddings,
        )
        self.json_utils = JsonUtils(
            app_name=self.model.app_name, app_author=self.model.app_author, file_name="rag_sources.json"
        )

    def chain(
        self,
        *,
        chat: BaseChatModel,
        chat_history: BaseChatMessageHistory,
        settings_model: SettingsModel,
    ):
        retriever = self.vectorstore.as_retriever()

        contextualize_q_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", settings_model.ai_rag_contextualize_prompt),
                MessagesPlaceholder("chat_history"),
                ("human", "{input}"),
            ]
        )
        history_aware_retriever = create_history_aware_retriever(chat, retriever, contextualize_q_prompt)

        qa_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", settings_model.ai_rag_system_prompt),
                MessagesPlaceholder("chat_history"),
                ("human", "{input}"),
            ]
        )
        question_answer_chain = create_stuff_documents_chain(chat, qa_prompt)

        question_answer_chain_w_history = RunnableWithMessageHistory(
            question_answer_chain,
            lambda session_id: chat_history,  # noqa: ARG005
            input_messages_key="input",
            history_messages_key="chat_history",
        )

        return create_retrieval_chain(history_aware_retriever, question_answer_chain_w_history)

    async def create(self, *, url: str, max_depth: int = 0):
        docs = await self.load_url_impl(url=url, max_depth=max_depth)
        ids = await self._load_documents(docs=docs)
        rag_sources: RAGSources = RAGSources()
        with contextlib.suppress(FileNotFoundError):
            rag_sources = await self.json_utils.read(rag_sources)

        rag_sources.rag_sources.append(
            RAGSource(
                rag_source_type=RAGSourceType.web_page,
                rag_source=url,
                rag_source_max_depth=max_depth,
                rag_document_ids=ids,
            )
        )
        await self.json_utils.write(rag_sources)

    async def read(self) -> RAGSources:
        rag_sources: RAGSources = RAGSources()
        with contextlib.suppress(FileNotFoundError):
            return await self.json_utils.read(rag_sources)
        return rag_sources

    async def delete(self, url: str):
        rag_sources: RAGSources = RAGSources()
        with contextlib.suppress(FileNotFoundError):
            rag_sources = await self.json_utils.read(rag_sources)

        new_rag_sources = []
        for rag_source in rag_sources.rag_sources:
            if rag_source.rag_source != url:
                rag_sources.append(rag_source)
            else:
                await self.vectorstore.adelete(ids=rag_source.rag_document_ids)

        rag_sources.rag_sources = new_rag_sources
        await self.json_utils.write(rag_sources)

    async def _load_documents(self, *, docs: Iterator[Document]) -> [str]:
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=500, chunk_overlap=0)
        split_docs = text_splitter.split_documents(docs)

        if self.model.vector_store_type == RAGVectorStoreType.CHROMA:
            filtered_docs = filter_complex_metadata(split_docs)
            return await self.vectorstore.aadd_documents(documents=filtered_docs)

        error_str = f"VectorStoreType {self.model.vector_store_type=} not implemented"
        raise NotImplementedError(error_str)

    async def load_url_impl(self, *, url: str, max_depth: int = 0) -> Iterator[Document]:
        def extractor(x):
            return Soup(x, "lxml")

        loader = RecursiveUrlLoader(
            url=url,
            max_depth=max_depth,
            extractor=lambda x: extractor(x).text,
            check_response_status=True,
        )
        return await loader.aload()

    @classmethod
    def create_vectorstore(
        cls,
        *,
        vector_store_type: RAGVectorStoreType,
        persist_directory: str,
        embeddings: Embeddings,
    ) -> VectorStore:
        if vector_store_type == RAGVectorStoreType.CHROMA:
            from chromadb.config import Settings

            return Chroma(
                persist_directory=persist_directory,
                embedding_function=embeddings,
                client_settings=Settings(
                    is_persistent=True,
                    allow_reset=True,
                ),
            )

        error_str = f"VectorStoreType {vector_store_type=} not implemented"
        raise NotImplementedError(error_str)



from loguru import logger as log
from textual.logging import TextualHandler

# log.add(
#     format="{time} {level} {file}:{line} - {message}",
# )

log.configure(
    handlers=[{"sink": TextualHandler(), "format": "{level} {file}:{line} - {message}"}],
)
# log.level("INFO")



import contextlib
import os
import shutil
import tempfile
from pathlib import Path

import pytest

from paita.utils import file_utils


@pytest.fixture
def folder():
    folder_path = tempfile.mkdtemp()
    with open(os.path.join(folder_path, "file1.txt"), "wb") as f:
        f.write(b"Hello")
    with open(os.path.join(folder_path, "file2.txt"), "wb") as f:
        f.write(b"World")

    yield folder_path
    with contextlib.suppress(FileNotFoundError):
        shutil.rmtree(folder_path)


@pytest.mark.asyncio
async def test_delete_folder(folder):
    await file_utils.delete_folder(Path(folder), recursive=True, dry_run=False)
    with pytest.raises(FileNotFoundError):
        shutil.rmtree(folder)



import pytest

from paita.llm.chat import AsyncHandler, Chat
from paita.llm.chat_history import ChatHistory
from paita.llm.enums import AIService
from paita.llm.models import get_embeddings
from paita.rag.rag_manager import RAGManager, RAGManagerModel, RAGVectorStoreType
from paita.utils.settings_model import SettingsModel

ai_service_models = {
    # AIService.AWSBedRock.value: "anthropic.claude-v2",
    # AIService.OpenAI.value: "gpt-3.5-turbo",
    AIService.Ollama.value: "llama3.1:latest",
}


@pytest.fixture(params=list(ai_service_models.items()))
def settings_model(request):
    key, value = request.param
    return SettingsModel(ai_service=key, ai_model=value)


@pytest.fixture(params=list(ai_service_models.items()))
def rag_manager(request):
    key, value = request.param
    return RAGManager(
        RAGManagerModel(
            app_name="test_rag_manager",
            app_author="unit_test_author",
            embeddings=get_embeddings(ai_service=key),
            vector_store_type=RAGVectorStoreType.CHROMA,
        )
    )


def mock_callback():
    pass


@pytest.fixture
def chat() -> Chat:
    return Chat()


@pytest.fixture
def callback_handler() -> AsyncHandler:
    callback_handler = AsyncHandler()
    callback_handler.register_callbacks(mock_callback, mock_callback, mock_callback)
    return callback_handler


@pytest.fixture
def mock_env(monkeypatch):
    monkeypatch.setenv("OPENAI_API_KEY", "SOMETHING")


@pytest.fixture
def chat_history():
    return ChatHistory(app_name="test", app_author="test", file_history=False)


@pytest.mark.integration
@pytest.mark.usefixtures("mock_env")
def test_init_models(chat, settings_model, chat_history, rag_manager, callback_handler):
    chat.init_model(
        settings_model=settings_model,
        chat_history=chat_history,
        rag_manager=rag_manager,
        callback_handler=callback_handler,
    )


@pytest.mark.integration
@pytest.mark.asyncio
async def test_request_empty_history(chat, settings_model, chat_history, rag_manager, callback_handler):
    chat.init_model(
        settings_model=settings_model,
        chat_history=chat_history,
        rag_manager=rag_manager,
        callback_handler=callback_handler,
    )

    await chat.request("First")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_request_some_history(chat, settings_model, chat_history, rag_manager, callback_handler):
    chat.init_model(
        settings_model=settings_model,
        chat_history=chat_history,
        rag_manager=rag_manager,
        callback_handler=callback_handler,
    )
    await chat.request("First")
    await chat.request("Second")



import pytest

from paita.llm.chat import AsyncHandler, Chat
from paita.llm.chat_history import ChatHistory
from paita.llm.enums import AIService
from paita.utils.settings_model import SettingsModel

ai_service_models = {
    AIService.AWSBedRock.value: "anthropic.claude-v2",
    # AIService.OpenAIChatGPT.value: "gpt-3.5-turbo",
}


@pytest.fixture(params=list(ai_service_models.items()))
def settings_model(request):
    key, value = request.param
    return SettingsModel(ai_service=key, ai_model=value)


def mock_callback():
    pass


@pytest.fixture
def chat() -> Chat:
    return Chat()


@pytest.fixture
def callback_handler() -> AsyncHandler:
    callback_handler = AsyncHandler()
    callback_handler.register_callbacks(mock_callback, mock_callback, mock_callback)
    return callback_handler


@pytest.fixture
def mock_env(monkeypatch):
    monkeypatch.setenv("OPENAI_API_KEY", "SOMETHING")


@pytest.fixture
def chat_history():
    return ChatHistory(app_name="test", app_author="test", file_history=False)



from __future__ import annotations

import json
import re
from typing import Any

import validators


def to_str(value: Any) -> str | None:
    return str(value)


def str_to_num(value: float | str) -> float | str:
    if isinstance(value, (float, int)):
        return value
    if isinstance(value, str):
        try:
            result = int(value)
        except ValueError:
            try:
                result = float(value)
            except ValueError:
                result = value
        return result
    msg = f"Invalid type {type(value)}"
    raise TypeError(msg)


def str_to_dict(value: str) -> dict:
    val = re.sub(r"\s+", "", value)

    result = {}
    for kv in val.split(","):
        if "=" in kv:
            key, value_str = kv.split("=")
            if not key or not value_str:
                msg = "Invalid key-value pair"
                raise ValueError(msg)
            result[key] = str_to_num(value_str)

    return result


def dict_to_str(value: dict) -> str:
    if not isinstance(value, dict):
        msg = "Input must be a dictionary"
        raise TypeError(msg)

    sorted_items = sorted(value.items())

    def is_nested(d):
        return any(isinstance(v, dict) for v in d.values())

    filtered_items = [(key, val) for key, val in sorted_items if not is_nested({key: val})]

    return ",".join(f"{key}={json.dumps(val)}" for key, val in filtered_items)


def fix_url(url: str, protocol: str = "http") -> str:
    if not url.startswith(("http://", "https://")):
        url = f"{protocol}://{url}"
    return url


def split_and_fix(source: str) -> [str]:
    pattern = r"[,\s;]+"
    urls = re.split(pattern, source)

    fixed_urls = []
    for url in urls:
        fixed_url = fix_url(url)
        if not validators.url(fixed_url):
            error_str = f"Invalid URL: {fixed_url}"
            raise ValueError(error_str)
        fixed_urls.append(fixed_url)

    return fixed_urls


def validate_and_fix_url(url: str) -> [str]:
    fixed_url = fix_url(url)
    if not validators.url(fixed_url):
        error_str = f"Invalid URL: {fixed_url}"
        raise ValueError(error_str)
    return fixed_url



from paita.llm.enums import Role
from paita.llm.message import Message


def test_message():
    # Initialize role with "str"
    message = Message(content="Hello", role=Role.question.value)
    assert message.content == "Hello"
    assert message.role == Role.question

    # Initialize role with enum
    message = Message(content="Hello", role=Role.question)
    assert message.content == "Hello"
    assert message.role == Role.question



import os.path
from enum import Enum
from pathlib import Path
from typing import List, Optional, Tuple

import aiofiles
from aiofiles import os as aiofiles_os
from appdirs import user_config_dir
from cache3 import DiskCache

from paita.llm.enums import Tag
from paita.utils.settings_model import SettingsModel


class SettingsBackendType(Enum):
    LOCAL = "local"
    AWS_APPCONFIG = "aws_appconfig"
    NONE = "none"


SETTINGS_FILE_NAME = "settings.json"


class SettingsManager:
    model: SettingsModel = None

    _backend_type: SettingsBackendType = None

    def __init__(
        self,
        *,
        model: Optional[SettingsModel] = None,
        cache: Optional[DiskCache] = None,
        app_name: str,
        app_author: str,
        backend_type: SettingsBackendType = SettingsBackendType.LOCAL,
    ):
        self._app_name: str = app_name
        self._app_author: str = app_author
        self._backend_type: SettingsBackendType = backend_type

        self.model: SettingsModel = model

        if not self.model:
            if not cache:
                err = "Cache not provided"
                raise ValueError(err)
            ai_services: List[Tuple[str, str]] = list(cache.keys(tag=Tag.AI_MODELS.value))
            ai_service = ai_services[0]

            ai_models: List[Tuple[str, str]] = cache.get(ai_service, [], tag=Tag.AI_MODELS.value)
            ai_model = ai_models[0]

            self.model = SettingsModel(
                ai_service=ai_service,
                ai_model=ai_model,
            )

    @classmethod
    async def load(
        cls,
        *,
        app_name: str,
        app_author: str,
        backend_type: SettingsBackendType = SettingsBackendType.LOCAL,
    ) -> "SettingsManager":
        if backend_type == SettingsBackendType.LOCAL:
            model: SettingsModel = await cls._load_and_parse(app_name=app_name, app_author=app_author)
            return SettingsManager(
                model=model,
                app_name=app_name,
                app_author=app_author,
                backend_type=backend_type,
            )
        raise NotImplementedError

    async def save(self):
        await self._dump_and_save(self.model, app_name=self._app_name, app_author=self._app_author)

    @classmethod
    async def delete(cls, *, app_name: str, app_author: str):
        config_dir = user_config_dir(appname=app_name, appauthor=app_author)
        file_path = Path(config_dir) / SETTINGS_FILE_NAME

        if file_path.exists():
            file_path.unlink(missing_ok=True)
            await aiofiles_os.rmdir(config_dir)

    @classmethod
    async def _load_and_parse(cls, *, app_name: str, app_author: str) -> SettingsModel:
        config_dir = user_config_dir(appname=app_name, appauthor=app_author)
        file_path = Path(config_dir) / SETTINGS_FILE_NAME
        async with aiofiles.open(file_path, "r") as json_file:
            json_data = await json_file.read()
            # json_obj = json.loads(json_data)
            model: SettingsModel = SettingsModel.model_validate_json(json_data)
            return model

    @classmethod
    async def _dump_and_save(cls, model: SettingsModel, *, app_name: str, app_author: str):
        config_dir = user_config_dir(appname=app_name, appauthor=app_author)
        file_path = Path(config_dir) / SETTINGS_FILE_NAME

        if not os.path.exists(config_dir):
            os.makedirs(config_dir)

        async with aiofiles.open(file_path, "w") as json_file:
            json_data = model.model_dump_json()
            # json_str = json.dumps(json_data)
            await json_file.write(json_data)



from pathlib import Path

from appdirs import user_config_dir


def compose_path(file_name: str, *, app_name: str, app_author: str) -> Path:
    config_dir = user_config_dir(appname=app_name, appauthor=app_author)
    return Path(config_dir) / file_name



from paita.llm.models import get_embeddings
from paita.rag.rag_manager import RAGManager, RAGManagerModel, RAGVectorStoreType
from paita.utils.settings_model import SettingsModel


def create_rag_manager(*, app_name: str, app_author: str, settings_model: SettingsModel) -> RAGManager:
    # TODO: Refactor, identical code with settings manager
    ai_service = settings_model.ai_service
    ai_model = settings_model.ai_model

    embeddings = get_embeddings(ai_service=ai_service, ai_model=ai_model)
    try:
        vector_store_type = RAGVectorStoreType(settings_model.ai_rag_vector_store_type)
    except ValueError:
        vector_store_type = RAGVectorStoreType.CHROMA
    return RAGManager(
        RAGManagerModel(
            app_name=app_name,
            app_author=app_author,
            vector_store_type=vector_store_type,
            embeddings=embeddings,
        )
    )



from langchain.callbacks.base import AsyncCallbackHandler, BaseCallbackHandler
from langchain.schema.output import LLMResult


class SyncHandler(BaseCallbackHandler):
    callback_on_token = None
    callback_on_end = None

    def register_callbacks(self, callback_on_token, callback_on_end):
        self.callback_on_token = callback_on_token
        self.callback_on_end = callback_on_end

    def on_llm_new_token(self, token: str) -> None:
        self.callback_on_token(token)

    def on_llm_end(self, response: LLMResult) -> None:
        self.callback_on_end(response)


class AsyncHandler(AsyncCallbackHandler):
    # callback_on_start = None
    callback_on_token = None
    callback_on_end = None
    callback_on_error = None

    def register_callbacks(
        self,
        # callback_on_start,
        callback_on_token,
        callback_on_end,
        callback_on_error,
    ):
        # self.callback_on_start = callback_on_start
        self.callback_on_token = callback_on_token
        self.callback_on_end = callback_on_end
        self.callback_on_error = callback_on_error

    # From base class
    # async def on_chat_model_start(
    #     self,
    #     serialized: Dict[str, Any],
    #     messages: List[List[BaseMessage]],
    #     *,
    #     run_id: UUID,
    #     parent_run_id: Optional[UUID] = None,
    #     tags: Optional[List[str]] = None,
    #     metadata: Optional[Dict[str, Any]] = None,
    #     **kwargs: Any,
    # ) -> Any:
    #     # log.debug(
    #     #     f"{serialized=} {messages=} {run_id=} {parent_run_id=} {tags=} {metadata=}"
    #     # )
    #     pass

    async def on_llm_new_token(
        self,
        token: str,
        # *,
        # chunk: Optional[Union[GenerationChunk, ChatGenerationChunk]] = None,
        # run_id: UUID,
        # parent_run_id: Optional[UUID] = None,
        # tags: Optional[List[str]] = None,
        # **kwargs: Any,
        **kwargs,  # noqa: ARG002
    ) -> None:
        # log.debug(f"{token=} {chunk=} {run_id=} {parent_run_id=} {tags=}")
        self.callback_on_token(token)

    async def on_llm_end(
        self,
        response: LLMResult,
        # *,
        # run_id: UUID,
        # parent_run_id: Optional[UUID] = None,
        # tags: Optional[List[str]] = None,
        # **kwargs: Any,
        **kwargs,  # noqa: ARG002
    ) -> None:
        # log.debug(f"{response=} {run_id=} {parent_run_id=} {tags=}")
        output = response.flatten().pop().generations.pop().pop().text
        self.callback_on_end(output)

    async def on_llm_error(
        self,
        error: BaseException,
        # *,
        # run_id: UUID,
        # parent_run_id: Optional[UUID] = None,
        # tags: Optional[List[str]] = None,
        # **kwargs: Any,
        **kwargs,  # noqa: ARG002
    ) -> None:
        # log.error(f"{error=} {run_id=} {parent_run_id=} {tags=}")
        self.callback_on_error(error)



# from textual import events
from textual.widgets import TextArea


class MultiLineInput(TextArea):
    def __init__(self, text="", *, id: str, multiline: bool = False, **kwargs):
        super().__init__(text, id=id, **kwargs)
        self.multiline = multiline

    # def on_key(self, event: events.Key) -> None:
    #     if self.multiline and event.key == "enter":
    #         self.text += "\n"
    #     else:
    #         super()._on_key(event)



from pathlib import Path
from shutil import rmtree

import pytest
from cache3 import DiskCache

from paita.llm.enums import AIService, Tag
from paita.tui.factory import create_rag_manager
from paita.utils.settings_model import SettingsModel

MODELS_A = ["modelA1", "modelA2", "modelA3"]
MODELS_B = ["modelB1", "modelB2", "modelB3"]


@pytest.fixture
def cache() -> DiskCache:
    directory = "paita unit-tests"
    path = Path(directory)
    if path.exists():
        rmtree(path.as_posix())

    _cache: DiskCache = DiskCache(directory)
    _cache.set(AIService.AWSBedRock.value, MODELS_A, tag=Tag.AI_MODELS.value)
    yield _cache

    if path.exists():
        rmtree(path.as_posix())


@pytest.mark.integration
def test_create_rag_manager():
    settings_model = SettingsModel(ai_service=AIService.AWSBedRock)
    rag_manager = create_rag_manager(
        app_name="test_rag_manager",
        app_author="unit_test_author",
        settings_model=settings_model,
    )
    assert rag_manager is not None



from pathlib import Path

from aiofiles import os as aiofiles_os

from paita.utils.logger import log


async def delete_folder(folder_path: Path, *, recursive: bool = False, dry_run: bool = False) -> None:
    if recursive:
        for file in await aiofiles_os.listdir(folder_path):
            path = folder_path / file
            is_file = await aiofiles_os.path.isfile(path)
            is_folder = await aiofiles_os.path.isdir(path)
            log.debug(f"{is_file=} {is_folder=}")
            if is_file:
                if not dry_run:
                    await aiofiles_os.remove(path)
            elif is_folder:
                await delete_folder(path, recursive=recursive, dry_run=dry_run)
            else:
                pass

    if not dry_run:
        await aiofiles_os.rmdir(folder_path)



from enum import Enum
from typing import List, Optional

from pydantic import BaseModel


class RAGSourceType(Enum):
    web_page = "web_page"


class RAGVectorStoreType(Enum):
    CHROMA = "chroma"


class RAGSource(BaseModel):
    rag_source_type: RAGSourceType = RAGSourceType.web_page
    rag_source: Optional[str] = None
    rag_source_max_depth: int = 1
    rag_document_ids: List[str] = []

    class Config:
        use_enum_values = False


class RAGSources(BaseModel):
    rag_sources: List[RAGSource] = []



from enum import Enum
from pathlib import PurePath
from typing import TYPE_CHECKING, Union

from appdirs import user_config_dir
from cache3 import DiskCache
from textual.app import App, ComposeResult, Widget
from textual.binding import Binding
from textual.containers import Container, Horizontal, VerticalScroll
from textual.widgets import Button, Footer, Header, Input, LoadingIndicator

from paita.llm.callbacks import AsyncHandler
from paita.llm.chat import Chat
from paita.llm.chat_history import ChatHistory
from paita.llm.enums import Tag
from paita.llm.models import list_all_models
from paita.localization import labels
from paita.tui.error_screen import ErrorScreen
from paita.tui.factory import create_rag_manager
from paita.tui.message_box import MessageBox
from paita.tui.multi_line_input import MultiLineInput
from paita.tui.settings_screen import SettingsScreen
from paita.tui.wait_screen import WaitScreen
from paita.utils.logger import log
from paita.utils.settings_manager import SettingsManager, SettingsModel

if TYPE_CHECKING:
    from paita.rag.rag_manager import RAGManager


class Role(Enum):
    QUESTION = "question"
    ANSWER = "answer"
    INFO = "info"
    ERROR = "error"


CACHE_DIR = user_config_dir(appname=labels.APP_TITLE, appauthor=labels.APP_AUTHOR)
CACHE_NAME = "cache"
CACHE_TTL = 24 * 60 * 60


TEXT_AREA = True


class ChatApp(App):
    TITLE = labels.APP_TITLE
    SUB_TITLE = labels.APP_SUBTITLE
    CSS_PATH = PurePath(__file__).parent / "styles" / "app.tcss"

    BINDINGS = [
        Binding("ctrl+q", "quit", "Quit", key_display="ctrl+q"),
        Binding("ctrl+x", "clear", "Clear", key_display="ctrl+x"),
        Binding("ctrl+1", "settings", "Settings", key_display="ctrl+1"),
    ]

    def __init__(self):
        super().__init__()

        self._settings_manager: SettingsManager = None
        self._rag_manager: RAGManager = None
        self._chat: Chat = None
        self._cache: DiskCache = DiskCache(CACHE_DIR, CACHE_NAME)

        self._current_message: Union[MessageBox or None] = None
        self._current_id: str = "id_0"
        self._last_focused: Union[Widget or None] = None

        self._chat_history = ChatHistory(app_name=labels.APP_TITLE, app_author=labels.APP_AUTHOR)

    def compose(self) -> ComposeResult:
        yield Header(show_clock=True)
        with Container(id="body"):
            vertical_scroll: VerticalScroll = VerticalScroll(id="conversation")
            vertical_scroll.can_focus = False
            yield vertical_scroll
            with Horizontal(id="input_box"):
                if TEXT_AREA:
                    yield MultiLineInput(id="multi_line_input", multiline=True)
                else:
                    yield Input(id="input")
                yield Button("Send", variant="success", id="send_button")
        yield Footer()

    # Action handlers

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "send_button":
            await self.process_conversation()

    async def on_input_submitted(self, event: Input.Submitted) -> None:
        if event.input.id in ("input", "multi_line_input"):
            await self.process_conversation()

    def action_settings(self, allow_cancel: bool = False) -> None:  # noqa: FBT001, FBT002
        settings_screen = SettingsScreen(
            settings_manager=self._settings_manager,
            rag_manager=self._rag_manager,
            cache=self._cache,
            allow_cancel=allow_cancel,
        )
        callback = self.exit_settings
        log.info(f"{callback=}")
        self.push_screen(
            screen=settings_screen,
            callback=callback,
        )

    async def action_clear(self) -> None:
        await self._chat_history.history.aclear()
        await self.query_one("#conversation").remove()
        await self.query_one("#body").mount(VerticalScroll(id="conversation"))

    def action_quit(self) -> None:
        self._cache.clear()
        self.exit()

    # Callbacks

    def exit_settings(self, changed: bool = False):  # noqa: FBT001, FBT002
        if changed:
            self.init_chat()

    async def on_mount(self) -> None:
        await self._mount_chat_history()
        try:
            await self.push_screen(WaitScreen(labels.APP_LIST_AI_SERVICES_MODELS))
            await self._list_models()
            self.pop_screen()
        except ValueError as e:
            log.error(e)
            self.pop_screen()
            await self.push_screen(
                ErrorScreen(
                    labels.APP_ERROR_NO_AI_SERVICES_OR_MODELS,
                    labels.APP_DIALOG_BUTTON_EXIT,
                ),
                self.exit,
            )
            return
        # Read existing settings or open settings screen
        try:
            self._settings_manager: SettingsManager = await SettingsManager.load(
                app_name=labels.APP_TITLE, app_author=labels.APP_AUTHOR
            )
            self._rag_manager = create_rag_manager(
                app_name=labels.APP_TITLE,
                app_author=labels.APP_AUTHOR,
                settings_model=self._settings_manager.model,
            )
            await self._rag_manager.read()
            self.init_chat()
        except FileNotFoundError:
            self._settings_manager = SettingsManager(
                cache=self._cache,
                app_name=labels.APP_TITLE,
                app_author=labels.APP_AUTHOR,
            )
            self._rag_manager = create_rag_manager(
                app_name=labels.APP_TITLE,
                app_author=labels.APP_AUTHOR,
                settings_model=self._settings_manager.model,
            )
            await self._rag_manager.read()
            self.action_settings(allow_cancel=False)

    def init_chat(self):
        if self._chat is None:
            self._chat = Chat()
        callback_handler = AsyncHandler()
        callback_handler.register_callbacks(self.callback_on_token, self.callback_on_end, self.callback_on_error)

        try:
            self._chat.init_model(
                settings_model=self._settings_manager.model,
                chat_history=self._chat_history,
                rag_manager=self._rag_manager,
                callback_handler=callback_handler,
            )
            if TEXT_AREA:
                self.query_one("#multi_line_input").focus()
            else:
                self.query_one("#input").focus()
        except ValueError:
            self._settings_manager = SettingsManager(
                model=SettingsModel(),
                app_name=labels.APP_TITLE,
                app_author=labels.APP_AUTHOR,
            )
            self.action_settings(allow_cancel=False)

    async def process_conversation(self) -> None:
        if TEXT_AREA:
            text_input: MultiLineInput = self.query_one("#multi_line_input")
            question = text_input.text
        else:
            text_input: Input = self.query_one("#input", Input)
            question = text_input.value

        if question == "":
            return

        button = self.query_one("#send_button")
        conversation = self.query_one("#conversation")

        self.toggle_widgets(text_input, button)

        with text_input.prevent(Input.Changed):
            if TEXT_AREA:
                text_input.text = ""
            else:
                text_input.value = ""

        await conversation.mount(
            MessageBox(question, role="question"),
            LoadingIndicator(),
        )
        conversation.scroll_end(animate=False)

        try:
            await self._chat.request(question)
        except ValueError as e:
            error = str(e)
            log.info(error)
            self.push_screen(ErrorScreen(error), self.exit_error_screen)
        except Exception as e:  # noqa: BLE001
            error = str(e)
            log.exception(e)
            self.push_screen(ErrorScreen(error), self.exit_error_screen)

    def toggle_widgets(self, *widgets: Widget) -> None:
        for w in widgets:
            w.disabled = not w.disabled

    def callback_on_token(self, data: str):
        if self._current_message is None:
            loading_indication = self.query_one(LoadingIndicator)
            loading_indication.remove()
            self._current_message = MessageBox(data, role="answer")
            conversation = self.query_one("#conversation")
            conversation.mount(self._current_message)
        else:
            self._current_message.append(data)

    def callback_on_end(self, data: str):
        conversation = self.query_one("#conversation")
        if self._current_message is None:
            loading_indication = self.query_one(LoadingIndicator)
            loading_indication.remove()
            self._current_message = MessageBox(data, role="answer")
            conversation.mount(self._current_message)

        self._current_message.flush()
        self._current_message = None

        text_input = self.query_one("#multi_line_input") if TEXT_AREA else self.query_one("#input")
        button = self.query_one("#send_button")
        self.toggle_widgets(text_input, button)
        text_input.focus()

    def callback_on_error(self, error):
        if self._current_message:
            loading_indication = self.query_one(LoadingIndicator)
            loading_indication.remove()
            self._current_message.flush()
            self._current_message = None

        self.push_screen(ErrorScreen(str(error)), self.exit_error_screen)

    def exit_error_screen(self, exit_app: bool = False):  # noqa: FBT001, FBT002
        if exit_app:
            self.exit()

        loading_indication = self.query_one(LoadingIndicator)
        loading_indication.remove()

        text_input = self.query_one("#multi_line_input") if TEXT_AREA else self.query_one("#input")
        button = self.query_one("#send_button")
        self.toggle_widgets(text_input, button)

        text_input.focus()

    async def _mount_chat_history(self):
        messages = await self._chat_history.messages()
        message_boxes: [MessageBox] = [
            MessageBox(data=message.content, role=message.role.value) for message in messages
        ]
        conversation = self.query_one("#conversation")
        await conversation.mount_all(message_boxes)
        conversation.scroll_end(animate=False)

    async def _list_models(self):
        self.push_screen(WaitScreen(labels.APP_LIST_AI_SERVICES_MODELS))
        all_models = await list_all_models()
        self.pop_screen()
        not_none_ai_models = any(len(models) for models in all_models.values())
        if not not_none_ai_models:
            msg = "No models found"
            raise ValueError(msg)
        for key in all_models:
            if all_models[key]:
                self._cache.set(key, all_models[key], CACHE_TTL, tag=Tag.AI_MODELS.value)


def main():
    app = ChatApp()
    app.run()


if __name__ == "__main__":
    main()



from pathlib import Path
from shutil import rmtree

import pytest
from cache3 import DiskCache

from paita.llm.enums import AIService, Tag

MODELS_A = ["modelA1", "modelA2", "modelA3"]
MODELS_B = ["modelB1", "modelB2", "modelB3"]


@pytest.fixture
def cache() -> DiskCache:
    directory = "paita unit-tests"
    path = Path(directory)
    if path.exists():
        rmtree(path.as_posix())

    _cache: DiskCache = DiskCache(directory)
    _cache.set(AIService.AWSBedRock.value, MODELS_A, tag=Tag.AI_MODELS.value)
    _cache.set(AIService.OpenAI.value, MODELS_B, tag=Tag.AI_MODELS.value)
    yield _cache

    if path.exists():
        rmtree(path.as_posix())


def test_get(cache):
    models_from_cache = cache.get(AIService.AWSBedRock.value, tag=Tag.AI_MODELS.value)
    assert models_from_cache == MODELS_A


def test_get_overwrite_existing(cache):
    cache.set(AIService.AWSBedRock.value, MODELS_B, tag=Tag.AI_MODELS.value)
    models_from_cache = cache.get(AIService.AWSBedRock.value, tag=Tag.AI_MODELS.value)
    assert models_from_cache == MODELS_B


def test_get_empty(cache):
    cache.clear()
    value = cache.get(AIService.AWSBedRock.value, tag=Tag.AI_MODELS.value)
    assert value is None


def test_keys(cache):
    ai_services = list(cache.keys(tag=Tag.AI_MODELS.value))
    assert sorted(ai_services) == sorted([AIService.AWSBedRock.value, AIService.OpenAI.value])



import os
from typing import Optional

from langchain_ollama import ChatOllama, OllamaEmbeddings
from ollama import AsyncClient

from paita.llm.services.service import Service
from paita.utils.logger import log


class Ollama(Service):
    @classmethod
    async def models(cls) -> [str]:
        try:
            client = AsyncClient(host=os.getenv("OLLAMA_ENDPOINT", None))
            response = await client.list()
            models = [model["name"] for model in response["models"]]
            return sorted(models)
        except Exception as e:  # noqa: BLE001 TODO
            log.info(e)
            return []

    @classmethod
    def embeddings(cls, model_id: Optional[str] = None) -> OllamaEmbeddings:
        return OllamaEmbeddings(model=model_id) if model_id else OllamaEmbeddings(model="llama3.1")

    def chat_model(self) -> ChatOllama:
        model_kwargs = {
            "max_tokens_to_sample": self._settings_model.ai_max_tokens,
            "temperature": 1,
            "top_k": 250,
            "top_p": 0.8,
            # "stop_sequences": ["\n\nHuman"],
        }
        if self._settings_model.ai_model_kwargs:
            model_kwargs.update(self._settings_model.ai_model_kwargs)
        log.debug(f"{model_kwargs=}")
        chat_ollama = ChatOllama(
            model=self._settings_model.ai_model,
            streaming=self._settings_model.ai_streaming,
            model_kwargs=model_kwargs,
            # max_tokens=settings_model.ai_max_tokens,
            # n=settings_model.ai_n,
            callbacks=[self._callback_handler],
            # callback_manager=callback_handler,
        )
        if (host := os.getenv("OLLAMA_ENDPOINT", None)) is not None:
            chat_ollama.base_url = host

        return chat_ollama



from typing import TYPE_CHECKING

from langchain_community.chat_message_histories import ChatMessageHistory, FileChatMessageHistory
from langchain_core.chat_history import AIMessage, BaseChatMessageHistory

from paita.llm.enums import Role
from paita.llm.message import Message
from paita.utils.config_dirs import compose_path

if TYPE_CHECKING:
    from pathlib import Path


HISTORY_FILE_NAME = "chat_history"


class ChatHistory:
    """
    ChatHistory is a factory for creating specific langchain ChatHistory instance
    """

    def __init__(self, *, app_name: str, app_author: str, file_history: bool = True):
        file_path: Path = compose_path(HISTORY_FILE_NAME, app_name=app_name, app_author=app_author)
        self.history: BaseChatMessageHistory = None
        if file_history:
            self.history = FileChatMessageHistory(str(file_path))
        else:
            self.history = ChatMessageHistory()

    async def messages(self) -> [Message]:
        # TODO: This conversion is quite unnecessary. Use Langchain classes directly.
        lc_messages = await self.history.aget_messages()
        messages: [Message] = []
        for lc_message in lc_messages:
            role = Role.question
            if type(lc_message) is AIMessage:
                role = Role.answer
            messages.append(Message(content=lc_message.content, role=role))
        return messages



from typing import TYPE_CHECKING

from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.schema.output_parser import StrOutputParser
from langchain_core.runnables.history import RunnableWithMessageHistory

from paita.llm.callbacks import AsyncHandler
from paita.llm.chat_history import ChatHistory
from paita.llm.models import AIService
from paita.llm.services import bedrock, ollama, openai
from paita.rag.rag_manager import RAGManager
from paita.utils.settings_model import SettingsModel

if TYPE_CHECKING:
    from langchain_core.language_models import BaseChatModel
    from langchain_core.runnables import Runnable


HISTORY_FILE_NAME = "chat_history"


class Chat:
    """
    Chat encapsulates chat history, RAG usage and can use different AI Models
    """

    def __init__(self):
        self._chat_model: BaseChatModel = None
        self._settings_model: SettingsModel = None
        self._chat_history: ChatHistory = None
        self._rag_manager: RAGManager = None
        self._chain: Runnable = None
        self._callback_handler: AsyncHandler = None
        self.parser: StrOutputParser = StrOutputParser()

    def init_model(
        self,
        *,
        settings_model: SettingsModel,
        chat_history: ChatHistory,
        rag_manager: RAGManager = None,
        callback_handler: AsyncHandler,
    ):
        self._settings_model = settings_model
        self._chat_history = chat_history
        self._rag_manager = rag_manager
        self._callback_handler = callback_handler

        if settings_model.ai_service == AIService.AWSBedRock.value:
            service = bedrock.Bedrock(settings_model=settings_model, callback_handler=self._callback_handler)
        elif settings_model.ai_service == AIService.OpenAI.value:
            service = openai.OpenAI(settings_model=settings_model, callback_handler=self._callback_handler)
        elif settings_model.ai_service == AIService.Ollama.value:
            service = ollama.Ollama(settings_model=settings_model, callback_handler=self._callback_handler)
        else:
            msg = f"Invalid AI Service {settings_model.ai_service}"
            raise ValueError(msg)
        self._chat_model = service.chat_model()
        if self._settings_model.ai_rag_enabled:
            self._chain = self._rag_manager.chain(
                chat=self._chat_model, chat_history=self._chat_history.history, settings_model=self._settings_model
            )
        else:
            prompt = ChatPromptTemplate.from_messages(
                [
                    (
                        "system",
                        self._settings_model.ai_persona,
                    ),
                    MessagesPlaceholder(variable_name="chat_history"),
                    ("human", "{input}"),
                ]
            )

            chain = prompt | self._chat_model | self.parser
            self._chain = RunnableWithMessageHistory(
                chain,
                lambda session_id: self._chat_history.history,  # noqa: ARG005
                input_messages_key="input",
                history_messages_key="chat_history",
            )

    async def request(self, data: str) -> str:
        await self._trim_history(self._chat_history, max_length=self._settings_model.ai_history_depth)

        if self._settings_model.ai_streaming:
            async for _ in self._chain.astream(
                {"input": data},
                {"configurable": {"session_id": "unused"}},
            ):
                pass
        else:
            await self._chain.ainvoke(
                {"input": data},
                {"configurable": {"session_id": "unused"}},
            )

    @classmethod
    async def _trim_history(cls, chat_history: ChatHistory, *, max_length: int = 20):
        stored_messages = await chat_history.history.aget_messages()
        if len(stored_messages) <= max_length:
            return

        await chat_history.history.aclear()
        trim_history_messages = stored_messages[-max_length:]
        await chat_history.history.aadd_messages(trim_history_messages)

    async def _summarize_messages(self, chat_history: ChatHistory, *, max_length: int = 20):
        stored_messages = chat_history.history.messages
        if len(stored_messages) <= max_length:
            return
        summarization_prompt = ChatPromptTemplate.from_messages(
            [
                MessagesPlaceholder(variable_name="chat_history"),
                (
                    "user",
                    "Distill the above chat messages into a single summary message."
                    "Include as many specific details as you can.",
                ),
            ]
        )
        summarization_chain = summarization_prompt | self._chat_model

        summary_message = summarization_chain.invoke({"chat_history": stored_messages})

        chat_history.history.clear()

        chat_history.history.add_message(summary_message)



import contextlib
from pathlib import Path
from typing import Type, TypeVar

import aiofiles
from aiofiles import os as aiofiles_os
from appdirs import user_config_dir
from pydantic import BaseModel

Model = TypeVar("Model", bound="BaseModel")


class JsonUtils:
    def __init__(self, *, app_name: str, app_author: str, file_name: str):
        self.app_name: str = app_name
        self.app_author: str = app_author
        self.config_dir: Path = Path(user_config_dir(appname=app_name, appauthor=app_author))
        self.file_name: Path = self.config_dir / file_name

    async def read(self, model: Type[Model]) -> Model:
        async with aiofiles.open(self.file_name, "r") as json_file:
            json_str = await json_file.read()
            return model.model_validate_json(json_str)
            # json_obj = json.loads(json_data)
            # log.debug(f"{json_obj=}")
            # model_instance = model.model_validate(json_obj)
            # log.debug(f"{model_instance=}")
            # return model_instance

    async def write(self, model: BaseModel):
        self.config_dir.mkdir(parents=True, exist_ok=True)

        async with aiofiles.open(self.file_name, "w") as json_file:
            json_str = model.model_dump_json()
            await json_file.write(json_str)

    async def delete(self):
        if self.file_name.exists():
            with contextlib.suppress(FileNotFoundError):
                self.file_name.unlink(missing_ok=True)
                await aiofiles_os.rmdir(self.config_dir)



from pathlib import PurePath
from typing import TYPE_CHECKING

import pyperclip
from rich.syntax import Syntax
from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal
from textual.widgets import Label, Markdown, Static
from textual.widgets._markdown import MarkdownBlock, MarkdownBullet, MarkdownFence

if TYPE_CHECKING:
    from textual.timer import Timer

ROLE_ABBREVIATIONS = {"question": "Q", "answer": "A", "info": "i", "error": "!"}


class MessageContent(Markdown, can_focus=True, can_focus_children=False):
    BINDINGS = [
        Binding("enter", "enable_focus_children"),
        Binding("escape", "disable_focus_children"),
        Binding("tab,down,right", "focus_next", priority=True),
        Binding("shift+tab,up,left", "focus_previous", priority=True),
        Binding("c", "copy", "Copy", show=True, key_display="c", priority=True),
    ]

    def action_enable_focus_children(self) -> None:
        self.can_focus_children = True
        for child in self.children:
            child.can_focus = True
        self.children[0].focus()

    def action_disable_focus_children(self) -> None:
        if isinstance(self.screen.focused, MarkdownFence):
            self.update_fence_highlighting(
                fence=self.screen.focused,
                highlight=False,
            )

        self.can_focus_children = False
        for child in self.children:
            child.can_focus = False
        self.focus()

    def action_focus_next(self) -> None:
        if self.can_focus_children:
            if isinstance(self.screen.focused, MarkdownFence):
                self.update_fence_highlighting(
                    fence=self.screen.focused,
                    highlight=False,
                )

            self.screen.focus_next("MessageContent *")
            if isinstance(self.screen.focused, MarkdownFence):
                self.update_fence_highlighting(fence=self.screen.focused, highlight=True)
        else:
            self.screen.focus_next()

    def action_focus_previous(self) -> None:
        if self.can_focus_children:
            if isinstance(self.screen.focused, MarkdownFence):
                self.update_fence_highlighting(
                    fence=self.screen.focused,
                    highlight=False,
                )

            self.screen.focus_previous("MessageContent *")
            if isinstance(self.screen.focused, MarkdownFence):
                self.update_fence_highlighting(self.screen.focused, highlight=True)
        else:
            self.screen.focus_previous()

    def update_fence_highlighting(
        self,
        fence: MarkdownFence,
        *,
        highlight: bool,
    ) -> None:
        if highlight:
            accent_color = "#0178D4"
            block = Syntax(
                fence.code,
                lexer=fence.lexer,
                word_wrap=False,
                indent_guides=True,
                padding=(1, 2),
                theme=fence.theme,
                background_color=accent_color,
            )
        else:
            block = Syntax(
                fence.code,
                lexer=fence.lexer,
                word_wrap=False,
                indent_guides=True,
                padding=(1, 2),
                theme=fence.theme,
            )

        fence.get_child_by_type(Static).update(block)

    def action_copy(self) -> None:
        if self.has_focus:
            content = self.parse_from_children(self)
        else:
            for child in self.children:
                if child.has_focus:
                    content = self.parse_from_children(child)

        pyperclip.copy(content)

    def parse_from_children(self, block: MarkdownBlock, indent_depth: int = -2) -> str:
        children = block.children

        if isinstance(block, MarkdownFence):
            return block.code

        if len(children) > 0:
            text = ""
            for child in children:
                text += self.parse_from_children(child, indent_depth + 1)
            return text
        if isinstance(block, MarkdownBullet):
            return " " * indent_depth + "- "
        if isinstance(block, MarkdownFence):
            return block.code
        return str(block.renderable) + "\n"


class MessageBox(Horizontal, can_focus=False):
    CSS_PATH = PurePath(__file__).parent / "styles" / "message_box.tcss"

    def __init__(self, data: str, *, role: str) -> None:
        super().__init__(classes=f"message {role}")
        self.data: str = data
        self._role: str = role
        self._message_content: MessageContent = None
        self._update_timer: Timer = None

    def compose(self) -> ComposeResult:
        role_label = ROLE_ABBREVIATIONS[self._role]
        yield Label(role_label, classes=f"{self._role}_label")
        self._message_content = MessageContent(self.data, classes="markdown")
        yield self._message_content

    def on_mount(self) -> None:
        self._update_timer = self.set_interval(1 / 4, self._markdown_update, pause=True)

    def append(self, data: str):
        self.data += data
        self._update_timer.resume()

    def flush(self):
        if self._update_timer:
            self._update_timer.pause()
        self._markdown_update()

    def _markdown_update(self):
        if self._message_content:
            self._message_content.update(self.data)
        self.parent.scroll_end(animate=False)



import contextlib
from pathlib import PurePath
from typing import List, Optional, Tuple

from cache3 import DiskCache
from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Horizontal, Vertical, VerticalScroll, Widget
from textual.css.query import NoMatches
from textual.screen import ModalScreen
from textual.validation import Function, Number
from textual.widgets import Button, Checkbox, Header, Input, Select, TextArea

import paita.localization.labels as label
from paita.llm.enums import Tag
from paita.localization import labels
from paita.rag.models import RAGSource, RAGSources
from paita.rag.rag_manager import RAGManager, RAGSourceType
from paita.tui.wait_screen import WaitScreen
from paita.utils.logger import log
from paita.utils.settings_manager import SettingsManager, SettingsModel
from paita.utils.string_utils import dict_to_str, str_to_dict, str_to_num, to_str, validate_and_fix_url

RAG_WIDGET_IDS = [
    "ai_rag_type",
    "ai_rag_source",
    "ai_rag_source_max_depth",
    "ai_rag_load",
    "ai_rag_reset",
    "ai_rag_contextualize_prompt",
    "ai_rag_system_prompt",
]


class SettingsScreen(ModalScreen[bool]):
    CSS_PATH = PurePath(__file__).parent / "styles" / "settings_screen.tcss"

    def __init__(
        self,
        *,
        settings_manager: SettingsManager,
        rag_manager: Optional[RAGManager],
        cache: DiskCache,
        allow_cancel: bool = False,
    ):
        super().__init__()
        self.settings_manager: SettingsManager = settings_manager
        self.rag_manager: RAGManager = rag_manager
        self.model: SettingsModel = settings_manager.model.model_copy()
        self.cache: DiskCache = cache
        self.allow_cancel: bool = allow_cancel

        # Prefill select options based on stored settings values
        self._ai_services: List[Tuple[str, str]] = list(self.cache.keys(tag=Tag.AI_MODELS.value))
        if self.model.ai_service not in self._ai_services:
            self.model.ai_service = self._ai_services[0]

        self._ai_models: List[Tuple[str, str]] = self.cache.get(self.model.ai_service, [], tag=Tag.AI_MODELS.value)
        if self.model.ai_model not in self._ai_models:
            self.model.ai_model = self._ai_models[0]

    def compose(self) -> ComposeResult:
        yield Header()
        with Container(classes="settings"):
            with VerticalScroll(id="settings"):
                with Horizontal(classes="settings_invisible_block"):
                    yield Select(
                        [(item, item) for item in self._ai_services],
                        value=self.model.ai_service,
                        prompt="AI Service",
                        id="ai_service",
                        classes="settings_small_option",
                        allow_blank=False,
                    )
                    yield Select(
                        [(item, item) for item in self._ai_models],
                        value=self.model.ai_model,
                        prompt="AI Model",
                        id="ai_model",
                        classes="settings_option",
                        allow_blank=False,
                    )
                    yield Checkbox(
                        label.AI_STREAMING,
                        value=self.model.ai_streaming,
                        id="ai_streaming",
                        classes="settings_checkbox",
                    )

                yield TextArea(
                    text=self.model.ai_persona,
                    id="ai_persona",
                    classes="settings_textarea",
                )

                with Horizontal(classes="settings_invisible_block"):
                    yield Input(
                        placeholder=label.AI_MODEL_KWARGS,
                        value=dict_to_str(self.model.ai_model_kwargs),
                        id="ai_model_kwargs",
                        classes="settings_input",
                        validators=[Function(self.validate_ai_model_kwargs, "Invalid dictionary")],
                    )
                    yield Input(
                        placeholder=label.AI_MAX_TOKENS,
                        value=to_str(self.model.ai_max_tokens),
                        id="ai_max_tokens",
                        classes="settings_small_input",
                        type="integer",
                        validators=[Number(minimum=1, maximum=99999)],
                        max_length=5,
                    )
                    yield Input(
                        placeholder=label.AI_HISTORY_DEPTH,
                        value=to_str(self.model.ai_history_depth),
                        id="ai_history_depth",
                        classes="settings_small_input",
                        type="integer",
                        validators=[Number(minimum=0, maximum=100)],
                        max_length=3,
                    )
                    # yield Input(
                    #     placeholder=label.AI_N,
                    #     value=to_str(self.model.ai_n),
                    #     id="ai_n",
                    #     classes="settings_input",
                    #     type="integer",
                    #     validators=[Number(minimum=1, maximum=10)],
                    #     max_length=4,
                    # )

                with Vertical(classes="settings_invisible_block", id="ai_rag_container"):
                    yield Checkbox(
                        label.AI_RAG_ENABLED,
                        value=self.model.ai_rag_enabled,
                        id="ai_rag_enabled",
                        classes="settings_checkbox",
                    )
                    yield TextArea(
                        text=self.model.ai_rag_contextualize_prompt,
                        id="ai_rag_contextualize_prompt",
                        classes="settings_textarea",
                    )
                    yield TextArea(
                        text=self.model.ai_rag_system_prompt,
                        id="ai_rag_system_prompt",
                        classes="settings_textarea",
                    )

                with Horizontal(classes="settings_block"):
                    yield Button(label="Apply", variant="success", id="apply")
                    if self.allow_cancel:
                        yield Button(label="Cancel", variant="warning", id="cancel")

    @staticmethod
    def validate_ai_model_kwargs(value: str):
        try:
            str_to_dict(value)
        except ValueError:
            return False
        else:
            return True

    async def on_mount(self) -> None:
        ai_rag_disabled = not self.model.ai_rag_enabled
        if self.model.ai_service is Select.BLANK or self.model.ai_model is Select.BLANK:
            self.query_one("#apply").disabled = True

            for rag_widget_id in RAG_WIDGET_IDS:
                with contextlib.suppress(NoMatches):
                    self.query_one(f"#{rag_widget_id}").disabled = True

            ai_rag_disabled = True
            self.query_one("#ai_rag_enabled").disabled = True

        # TODO: Create a separate component to handle RAG source(s)
        self.query_one("#ai_rag_contextualize_prompt").disabled = ai_rag_disabled
        self.query_one("#ai_rag_system_prompt").disabled = ai_rag_disabled

        rag_sources: RAGSources = await self.rag_manager.read()
        rag_source: RAGSource = rag_sources.rag_sources[-1] if rag_sources.rag_sources else RAGSource()

        widgets: [Widget] = [
            Select(
                [(item.value, item.value) for item in RAGSourceType],
                value=rag_source.rag_source_type.value,
                prompt="AI Rag Type",
                id="ai_rag_type",
                classes="settings_small_option",
                allow_blank=False,
                disabled=ai_rag_disabled,
            ),
            Input(
                placeholder=label.AI_RAG_SOURCE_MAX_DEPTH,
                value=str(rag_source.rag_source_max_depth),
                id="ai_rag_source_max_depth",
                classes="settings_small_input",
                type="integer",
                validators=[Number(minimum=1, maximum=5)],
                max_length=1,
                disabled=ai_rag_disabled,
            ),
        ]

        # ai_rag_source: not None
        # - input = disable
        # - button = "reset"
        # ai_rag_source: None
        # - input = enabled
        # - button = "load"
        input_disabled = ai_rag_disabled or bool(rag_source.rag_source)
        ai_rag_source = Input(
            placeholder=label.AI_RAG_SOURCE,
            value=rag_source.rag_source if rag_source.rag_source else "",
            id="ai_rag_source",
            classes="settings_input",
            disabled=input_disabled,
        )
        widgets.append(ai_rag_source)

        button_disabled = ai_rag_disabled
        if not ai_rag_disabled and ai_rag_source.value == "":
            button_disabled = True
        if bool(rag_source.rag_source):
            widgets.append(Button(label="Reset", variant="warning", id="ai_rag_reset", disabled=button_disabled))
        else:
            widgets.append(Button(label="Load", variant="success", id="ai_rag_load", disabled=button_disabled))

        horizontal = Horizontal(classes="settings_invisible_block", id="ai_rag_sources")
        rag_container = self.query_one("#ai_rag_container")
        await rag_container.mount(horizontal)
        await horizontal.mount(*widgets)
        self.query_one("#ai_persona").disabled = self.model.ai_rag_enabled

    @on(Input.Changed)
    async def input_changed(self, event: Input.Changed) -> None:
        if event.input.id == "ai_rag_source":
            try:
                ai_rag_load = self.query_one("#ai_rag_load")
                if event.input.value != "":
                    ai_rag_load.disabled = False
                else:
                    ai_rag_load.disabled = True
            except NoMatches:
                pass

    @on(Checkbox.Changed)
    async def checkbox_changed(self, event: Checkbox.Changed) -> None:
        if event.checkbox.id == "ai_rag_enabled":
            for rag_widget_id in RAG_WIDGET_IDS:
                with contextlib.suppress(NoMatches):
                    self.query_one(f"#{rag_widget_id}").disabled = not event.checkbox.value

            # Keep 'rag source'-input field disabled when we think source is loaded
            # If there's 'load'-button keep it disabled if 'rag source' field has content
            if event.checkbox.value:
                ai_rag_source = self.query_one("#ai_rag_source")
                try:
                    ai_rag_load = self.query_one("#ai_rag_load")
                    if ai_rag_source.value != "":
                        ai_rag_load.disabled = False
                    else:
                        ai_rag_load.disabled = True
                except NoMatches:
                    pass

            self.query_one("#ai_persona").disabled = event.checkbox.value

    @on(Select.Changed)
    async def select_changed(self, event: Select.Changed) -> None:
        if event.control.id == "ai_service":
            value = event.value
            if value == self.model.ai_service:
                return
            self.settings_manager.ai_service = value

            models = self.cache.get(event.value, Select.BLANK, tag=Tag.AI_MODELS.value)
            widget: Select = self.query_one("#ai_model")
            widget.set_options((item, item) for item in models)
        elif event.control.id == "ai_model":
            value = event.value
            if value == self.model.ai_model:
                return
            self.query_one("#apply").disabled = False
        elif event.control.id == "ai_rag_type":
            pass  # TODO: Handle different RAG types
        else:
            log.error(f"Undefined {event.control.id=}")

    async def on_button_pressed(self, event: Button.Pressed) -> None:
        # PoC RAG approach, we always update model and save
        model: SettingsModel = SettingsModel(
            ai_service=self.query_one("#ai_service").value,
            ai_model=self.query_one("#ai_model").value,
        )
        if (value := self.query_one("#ai_persona").text) != "":
            model.ai_persona = value
        if (value := self.query_one("#ai_model_kwargs").value) != "":
            model.ai_model_kwargs = str_to_dict(value)
        model.ai_streaming = self.query_one("#ai_streaming").value
        # if (value := self.query_one("#ai_n").value) != "":
        #     model.ai_n = str_to_num(value)
        if (value := self.query_one("#ai_max_tokens").value) != "":
            model.ai_max_tokens = str_to_num(value)
        if (value := self.query_one("#ai_history_depth").value) != "":
            model.ai_history_depth = str_to_num(value)

        # RAG
        model.ai_rag_enabled = self.query_one("#ai_rag_enabled").value
        if (value := self.query_one("#ai_rag_contextualize_prompt").text) != "":
            model.ai_rag_contextualize_prompt = value
        if (value := self.query_one("#ai_rag_system_prompt").text) != "":
            model.ai_rag_system_prompt = value

        self.settings_manager.model = model

        if event.button.id == "apply":
            await self.settings_manager.save()
            self.dismiss(True)
        elif event.control.id == "ai_rag_load":
            ai_rag_source = self.query_one("#ai_rag_source")
            if (value := ai_rag_source.value) != "":
                try:
                    url = validate_and_fix_url(value)
                    ai_rag_source.remove_class("settings_input_error", update=True)
                    ai_rag_source.value = url

                    max_depth = str_to_num(self.query_one("#ai_rag_source_max_depth").value)
                    await self.parent.push_screen(WaitScreen(labels.APP_RAG_PROCESS_DOCUMENTS))
                    await self.rag_manager.create(url=url, max_depth=max_depth)
                    await event.control.remove()
                    await self.mount(Button(label="Reset", variant="warning", id="ai_rag_reset"), after=ai_rag_source)

                    await self.settings_manager.save()
                except ValueError:
                    ai_rag_source.add_class("settings_input_error", update=True)

                self.parent.pop_screen()
        elif event.control.id == "ai_rag_reset":
            ai_rag_source = self.query_one("#ai_rag_source")
            await self.rag_manager.delete(ai_rag_source.value)
            ai_rag_source.value = ""
            ai_rag_source.disabled = False
            await event.control.remove()
            await self.mount(
                Button(label="Load", variant="success", id="ai_rag_load", disabled=True), after=ai_rag_source
            )
            await self.settings_manager.save()
        else:
            self.dismiss(False)



import pytest

from paita.utils.string_utils import dict_to_str, split_and_fix, str_to_dict, str_to_num


def test_to_num():
    assert str_to_num("1") == 1
    assert str_to_num("2.1") == 2.1
    assert str_to_num("string") == "string"
    assert str_to_num(1) == 1
    assert str_to_num(2.1) == 2.1


def test_str_to_dict():
    value_str = " temperature=1, top_p= 0.95 ,\t max_tokens_to_sample = 1000\n"
    value_dict = str_to_dict(value_str)
    assert value_dict == {
        "temperature": 1,
        "top_p": 0.95,
        "max_tokens_to_sample": 1000,
    }


def test_str_to_dict_invalid():
    invalid_1 = " temperature, top_p= 0.95 ,\t max_tokens_to_sample = 1000\n"
    invalid_2 = " temperature=1, top_p= ,\t max_tokens_to_sample = 1000\n"
    invalid_3 = " temperature=1, top_p ,\t max_tokens_to_sample = 1000\n"
    invalid_4 = " temperature=1,,\t max_tokens_to_sample = 1000\n"

    assert str_to_dict(invalid_1) == {
        "max_tokens_to_sample": 1000,
        "top_p": 0.95,
    }

    with pytest.raises(ValueError, match="Invalid key-value pair"):
        str_to_dict(invalid_2)

    assert str_to_dict(invalid_3) == {
        "max_tokens_to_sample": 1000,
        "temperature": 1,
    }

    assert str_to_dict(invalid_4) == {
        "max_tokens_to_sample": 1000,
        "temperature": 1,
    }


def test_dict_to_str():
    value_dict = {
        "temperature": 1,
        "top_p": 0.95,
        "max_tokens_to_sample": 1000,
        "discard": {"nested": "dictionary"},
    }
    value_str = dict_to_str(value_dict)
    assert value_str == "max_tokens_to_sample=1000,temperature=1,top_p=0.95"


def test_split_and_validate():
    sources = "http://aws.amazon.com, https://openai.com, https://ollama.com"
    output = split_and_fix(sources)
    assert output == ["http://aws.amazon.com", "https://openai.com", "https://ollama.com"]


def test_split_and_validate_invalid():
    sources = "aws.amazon.com, openai.com, ollama.com"
    split_sources = split_and_fix(sources)
    assert len(split_sources) == 3
    assert split_sources[0] == "http://aws.amazon.com"



from typing import Optional

from langchain_core.embeddings import Embeddings
from langchain_core.language_models.chat_models import BaseChatModel

from paita.llm.callbacks import AsyncHandler
from paita.utils.settings_model import SettingsModel


class Service:
    def __init__(self, *, settings_model: SettingsModel, callback_handler: AsyncHandler):
        self._settings_model = settings_model
        self._callback_handler = callback_handler

    @classmethod
    async def models(cls) -> [str]:
        raise NotImplementedError

    @classmethod
    def embeddings(cls, model_id: Optional[str] = None) -> Embeddings:
        raise NotImplementedError

    def chat_model(self) -> BaseChatModel:
        raise NotImplementedError



from enum import Enum


class AIService(Enum):
    AWSBedRock = "AWS Bedrock"
    OpenAI = "OpenAI"
    Ollama = "Ollama"


class Tag(Enum):
    AI_SERVICE = "ai_services"
    AI_MODELS = "ai_models"


class Role(Enum):
    question = "question"
    answer = "answer"
    info = "info"



import pytest

from paita.llm.enums import AIService
from paita.utils.settings_manager import SettingsManager, SettingsModel


@pytest.mark.asyncio
async def test_load_unsupported_backend():
    with pytest.raises(NotImplementedError):
        await SettingsManager.load(
            app_name="paita_unit_tests",
            app_author="unit_test",
            backend_type="some_unsupported",
        )


@pytest.mark.asyncio
async def test_save_load_local():
    model: SettingsModel = SettingsModel(
        version=1.0,
        ai_service=AIService.AWSBedRock.value,
        ai_model="SomeModel",
        ai_persona="Custom Persona",
    )
    manager: SettingsManager = SettingsManager(
        model=model,
        app_name="paita_unit_tests",
        app_author="unit_test",
    )
    try:
        await manager.save()
        loaded_manager: SettingsManager = await SettingsManager.load(
            app_name="paita_unit_tests", app_author="unit_test"
        )
        assert loaded_manager.model == model
    finally:
        await SettingsManager.delete(app_name="paita_unit_tests", app_author="unit_test")



import pytest
from pydantic import ValidationError

from paita.llm.enums import AIService
from paita.llm.models import get_embeddings
from paita.rag.models import RAGSource, RAGSources, RAGSourceType
from paita.rag.rag_manager import RAGManager, RAGManagerModel, RAGVectorStoreType


@pytest.fixture
def rag_manager() -> RAGManager:
    return RAGManager(
        RAGManagerModel(
            app_name="test_rag_manager",
            app_author="unit_test_author",
            embeddings=get_embeddings(ai_service=AIService.AWSBedRock.value),
            vector_store_type=RAGVectorStoreType.CHROMA,
        )
    )


@pytest.mark.integration
def test_rag_manager(rag_manager: RAGManager):
    assert rag_manager is not None


@pytest.mark.integration
def test_rag_manager_invalid():
    with pytest.raises(ValidationError):
        return RAGManager(
            RAGManagerModel(
                app_name=None,
                app_author="unit_test_author",
                embeddings=get_embeddings(ai_service=AIService.AWSBedRock.value),
                vector_store_type=RAGVectorStoreType.CHROMA,
            )
        )

    with pytest.raises(ValidationError):
        return RAGManager(
            RAGManagerModel(
                app_name="test_rag_manager",
                app_author="unit_test_author",
                embeddings=get_embeddings(ai_service=AIService.AWSBedRock.value),
                vector_store_type="something",
            )
        )


@pytest.mark.integration
@pytest.mark.asyncio
async def test_load_url_impl(rag_manager: RAGManager):
    url = "https://www.google.com"
    docs = await rag_manager.load_url_impl(url=url, max_depth=1)
    assert docs is not None


@pytest.mark.integration
@pytest.mark.asyncio
async def test_create(rag_manager: RAGManager):
    url = "https://www.google.com"
    await rag_manager.create(url=url, max_depth=1)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_read_na(rag_manager: RAGManager):
    rag_sources = await rag_manager.read()
    assert rag_sources is not None


@pytest.mark.integration
@pytest.mark.asyncio
async def test_read(rag_manager: RAGManager):
    try:
        await rag_manager.create(url="https://www.google.com", max_depth=1)
        rag_sources: RAGSources = await rag_manager.read()
        assert rag_sources is not None
        assert len(rag_sources.rag_sources) > 0
        rag_source: RAGSource = rag_sources.rag_sources[0]
        assert type(rag_source.rag_source_type) is RAGSourceType
    finally:
        pass
        # await delete_folder(rag_manager.file_path, recursive=True)


@pytest.mark.integration
@pytest.mark.asyncio
async def test_delete(rag_manager: RAGManager):
    await rag_manager.create(url="https://www.google.com", max_depth=1)
    await rag_manager.delete("https://www.google.com")



from __future__ import annotations

import asyncio
import os
from functools import partial
from typing import Optional

import boto3
from botocore.config import Config
from langchain_aws import BedrockEmbeddings, ChatBedrock

from paita.llm.services.service import Service
from paita.utils.logger import log


class Bedrock(Service):
    @classmethod
    async def models(cls):
        loop = asyncio.get_event_loop()
        try:
            bedrock_client = cls._get_bedrock_client(runtime=False)
            pf = partial(bedrock_client.list_foundation_models, byOutputModality="TEXT")
            response = await loop.run_in_executor(None, pf)
            models = [model["modelId"] for model in response["modelSummaries"]]
            return sorted(models)
        except Exception as e:  # noqa: BLE001 TODO
            log.info(e)
            return []

    @classmethod
    def embeddings(cls, model_id: Optional[str] = None) -> BedrockEmbeddings:
        return BedrockEmbeddings(model_id=model_id) if model_id else BedrockEmbeddings()

    def chat_model(self) -> ChatBedrock:
        model_kwargs = {
            # "max_tokens_to_sample": self._settings_model.ai_max_tokens,
            "temperature": 1,
            "top_k": 250,
            "top_p": 0.8,
            # "stop_sequences": ["\n\nHuman"],
        }
        if self._settings_model.ai_model_kwargs:
            model_kwargs.update(self._settings_model.ai_model_kwargs)
        log.debug(f"{model_kwargs=}")
        return ChatBedrock(
            model_id=self._settings_model.ai_model,
            streaming=self._settings_model.ai_streaming,
            model_kwargs=model_kwargs,
            # max_tokens=settings_model.ai_max_tokens,
            # n=settings_model.ai_n,
            callbacks=[self._callback_handler],
            # callback_manager=callback_handler,
        )

    @classmethod
    def _get_bedrock_client(
        cls,
        *,
        assumed_role: Optional[str] = None,
        region: Optional[str] = None,
        runtime: Optional[bool] = True,
    ):
        target_region = os.environ.get("AWS_REGION", os.environ.get("AWS_DEFAULT_REGION")) if region is None else region

        session_kwargs = {"region_name": target_region}
        client_kwargs = {**session_kwargs}

        profile_name = os.environ.get("AWS_PROFILE")
        if profile_name:
            session_kwargs["profile_name"] = profile_name

        retry_config = Config(
            region_name=target_region,
            connect_timeout=10,
            read_timeout=3,
            retries={
                "max_attempts": 3,
                "mode": "standard",
            },
        )
        session = boto3.Session(**session_kwargs)

        if assumed_role:
            sts = session.client("sts")
            response = sts.assume_role(RoleArn=str(assumed_role), RoleSessionName="langchain-llm-1")
            client_kwargs["aws_access_key_id"] = response["Credentials"]["AccessKeyId"]
            client_kwargs["aws_secret_access_key"] = response["Credentials"]["SecretAccessKey"]
            client_kwargs["aws_session_token"] = response["Credentials"]["SessionToken"]
        service_name = "bedrock-runtime" if runtime else "bedrock"

        return session.client(service_name=service_name, config=retry_config, **client_kwargs)



APP_TITLE = "paita"
APP_SUBTITLE = "Python AI Textual Assistant"
APP_AUTHOR = "paita"
AI_PERSONA_PREFIX = "AI Persona:"
AI_STREAMING = "Streaming"
AI_MODEL_KWARGS = "Extra arguments to pass to model"
AI_MAX_TOKENS = "Max tokens"
AI_HISTORY_DEPTH = "History depth"
AI_N = "Number of response messages"
AI_RAG_ENABLED = "RAG Enabled"
AI_RAG_SOURCE = "Source"
AI_RAG_SOURCE_MAX_DEPTH = "Max depth"

APP_LIST_AI_SERVICES_MODELS = "Checking available AI Services and AI Models"

APP_RAG_PROCESS_DOCUMENTS = "Processing documents..."
APP_ERROR_NO_AI_SERVICES_OR_MODELS = "No available AI Services or AI Models founds"

APP_DIALOG_BUTTON_EXIT = "Exit"



import asyncio
from typing import TYPE_CHECKING, Optional

from langchain_openai import ChatOpenAI
from langchain_openai.embeddings import OpenAIEmbeddings
from openai import OpenAI as OpenAIModule

from paita.llm.services.service import Service
from paita.utils.logger import log

if TYPE_CHECKING:
    from openai.types import Model


class OpenAI(Service):
    @classmethod
    async def models(cls):
        loop = asyncio.get_event_loop()
        try:
            client = OpenAIModule()
            pf = client.models.list
            response: [Model] = await loop.run_in_executor(None, pf)
            models = [model.id for model in response.data]
            return sorted(models)
        except Exception as e:  # noqa: BLE001 TODO
            log.info(e)
            return []

    @classmethod
    def embeddings(cls, model_id: Optional[str] = None) -> OpenAIEmbeddings:
        return OpenAIEmbeddings(model=model_id) if model_id else OpenAIEmbeddings()

    def chat_model(self) -> ChatOpenAI:
        model_kwargs = {
            # "top_k": 250,
            "top_p": 0.8,
            "frequency_penalty": 0.8,
            "presence_penalty": 0.8,
            # "stop_sequences": ["\n\nHuman"],
        }
        temperature = 1
        if self._settings_model.ai_model_kwargs is not None:
            model_kwargs.update(self._settings_model.ai_model_kwargs)
            if "temperature" in model_kwargs:
                temperature = int(model_kwargs["temperature"])
                del model_kwargs["temperature"]
        log.debug(f"{model_kwargs=}")
        return ChatOpenAI(
            model_name=self._settings_model.ai_model,
            streaming=self._settings_model.ai_streaming,
            model_kwargs=model_kwargs,
            max_tokens=self._settings_model.ai_max_tokens,
            temperature=temperature,
            # n=settings_model.ai_n,
            callbacks=[self._callback_handler],
            # callback_manager=callback_handler,
        )



from pathlib import PurePath

from textual.app import ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.screen import ModalScreen
from textual.widgets import Button, Label


class ErrorScreen(ModalScreen[bool]):
    CSS_PATH = PurePath(__file__).parent / "styles" / "error_screen.tcss"

    def __init__(self, error: str, button_text: str = "Ok"):
        super().__init__()
        self._error: str = error
        self._button_text: str = button_text

    def compose(self) -> ComposeResult:
        with VerticalScroll(id="error_screen_vertical"):
            yield Label(self._error, id="error_screen_label")
            with Horizontal(id="error_screen_button_block"):
                yield Button(self._button_text, variant="primary", id="error_screen_button")

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "error_screen_button":
            self.dismiss(False)



import asyncio
from typing import Dict, List, Optional

from paita.llm.enums import AIService
from paita.llm.services import bedrock, ollama, openai
from paita.utils.logger import log


async def list_models(ai_service: str):
    if ai_service == AIService.AWSBedRock.value:
        return await bedrock.Bedrock.models()
    if ai_service == AIService.OpenAI.value:
        return await openai.OpenAI.models()
    if ai_service == AIService.Ollama.value:
        return await ollama.Ollama.models()
    # if ai_service == AIService.Mock:
    #     return [
    #         "mock_model_1",
    #         "mock_model_2"
    #     ]
    msg = f"Invalid value for {ai_service=}"
    raise ValueError(msg)


def get_embeddings(*, ai_service: str, ai_model: Optional[str] = None):
    if ai_service == AIService.AWSBedRock.value:
        return bedrock.Bedrock.embeddings(ai_model)
    if ai_service == AIService.OpenAI.value:
        return openai.OpenAI.embeddings(ai_model)
    if ai_service == AIService.Ollama.value:
        return ollama.Ollama.embeddings(ai_model)
    # if ai_service == AIService.Mock:
    #     return [
    #         "mock_model_1",
    #         "mock_model_2"
    #     ]
    msg = f"Invalid value for {ai_service=}"
    raise ValueError(msg)


async def list_all_models() -> Dict[str, List[str]]:
    services = [service.value for service in AIService]
    tasks = [list_models(service) for service in services]
    responses: [[str]] = await asyncio.gather(*tasks)
    response: Dict[str, List[str]] = dict(zip(services, responses))
    return response


async def main():
    log.debug(await bedrock.Bedrock.models())
    log.debug(await ollama.Ollama.models())
    log.debug(await openai.OpenAI.models())
    log.debug(await list_all_models())


if __name__ == "__main__":
    asyncio.run(main())



from pathlib import PurePath

from textual.app import ComposeResult
from textual.containers import Vertical
from textual.screen import ModalScreen
from textual.widgets import Label, LoadingIndicator


class WaitScreen(ModalScreen[bool]):
    CSS_PATH = PurePath(__file__).parent / "styles" / "wait_screen.tcss"

    def __init__(self, text):
        super().__init__()
        self._text = text

    def compose(self) -> ComposeResult:
        with Vertical(id="wait_screen_vertical"):
            yield Label(self._text, id="wait_screen_label")
            yield LoadingIndicator()
